/*
        bzip2dll
                Copyright 1999 by Christian Ghisler (support@ghisler.com)
                This file is Freeware

        to be used with Windows Commander (www.ghisler.com)
*/


#define BZ_EXPORT           /* important: do not link to DLL! */
#include "bzlib.h"
#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>

#ifdef _MSC_VER
#include <sys/types.h>
#include <sys/stat.h>
#include <time.h>
#endif

#ifdef _WIN32
#include <io.h>
#include <dos.h>
#endif

char inifilename[MAX_PATH]="wcplugin.ini";  // Unused in this plugin, may be used to save data

/* Error codes returned to calling application */
#define E_END_ARCHIVE     10            /* No more files in archive */
#define E_NO_MEMORY       11            /* Not enough memory */
#define E_BAD_DATA        12            /* Data is bad */
#define E_BAD_ARCHIVE     13            /* CRC error in archive data */
#define E_UNKNOWN_FORMAT  14            /* Archive format unknown */
#define E_EOPEN           15            /* Cannot open existing file */
#define E_ECREATE         16            /* Cannot create file */
#define E_ECLOSE          17            /* Error closing file */
#define E_EREAD           18            /* Error reading from file */
#define E_EWRITE          19            /* Error writing to file */
#define E_SMALL_BUF       20            /* Buffer too small */
#define E_EABORTED        21            /* Function aborted by user */
#define E_NO_FILES        22            /* No files found */
#define E_TOO_MANY_FILES  23            /* Too many files to pack */
#define E_NOT_SUPPORTED   24            /* Function not supported */

/* flags for unpacking */
#define PK_OM_LIST         0
#define PK_OM_EXTRACT      1

/* flags for ProcessFile */
#define PK_SKIP            0            /* Skip this file */
#define PK_TEST            1            /* Test file integrity */
#define PK_EXTRACT         2            /* Extract to disk */

/* Flags passed through ChangeVolProc */
#define PK_VOL_ASK         0            /* Ask user for location of next volume */
#define PK_VOL_NOTIFY      1            /* Notify app that next volume will be unpacked */

/* Flags for packing */

/* For PackFiles */
#define PK_PACK_MOVE_FILES 1    /* Delete original after packing        */
#define PK_PACK_SAVE_PATHS 2    /* Save path names of files             */

/* Returned by GetPackCaps */
#define PK_CAPS_NEW        1    /* Can create new archives              */
#define PK_CAPS_MODIFY     2    /* Can modify exisiting archives        */
#define PK_CAPS_MULTIPLE   4    /* Archive can contain multiple files   */
#define PK_CAPS_DELETE     8    /* Can delete files                     */
#define PK_CAPS_OPTIONS   16    /* Has options dialog                   */
#define PK_CAPS_MEMPACK   32    /* Functions to pack in memory          */
#define PK_CAPS_BY_CONTENT 64   /* Can detect archive by its contents   */
#define PK_CAPS_SEARCHTEXT 128  /* Allow searching in this archive type */

#define BACKGROUND_UNPACK 1      /* Which operations are thread-safe?    */
#define BACKGROUND_PACK 2
#define BACKGROUND_MEMPACK 4

/* Options for memory packing */
#define MEM_OPTIONS_WANTHEADERS 1

/* Returned by PackToMem */
#define MEMPACK_OK         0
#define MEMPACK_DONE       1
#define MEMPACK_ERR       -1

typedef struct {
    char ArcName[260];
    char FileName[260];
    int Flags;
    int PackSize;
    int UnpSize; 
    int HostOS;
    int FileCRC;
    int FileTime;
    int UnpVer;
    int Method;
    int FileAttr;
    char* CmtBuf;
    int CmtBufSize;
    int CmtSize;
    int CmtState;
  } tHeaderData;

typedef struct {
    char ArcName[1024];
    char FileName[1024];
    int Flags;
    DWORD PackSize;
    int PackSizeHigh;
    DWORD UnpSize; 
    int UnpSizeHigh; 
    int HostOS;
    int FileCRC;
    int FileTime;
    int UnpVer;
    int Method;
    int FileAttr;
    char* CmtBuf;
    int CmtBufSize;
    int CmtSize;
    int CmtState;
    char Unused[1024];
  } tHeaderDataEx;

typedef struct {
    WCHAR ArcNameW[1024];
    WCHAR FileNameW[1024];
    int Flags;
    DWORD PackSize;
    int PackSizeHigh;
    DWORD UnpSize; 
    int UnpSizeHigh; 
    int HostOS;
    int FileCRC;
    int FileTime;
    int UnpVer;
    int Method;
    int FileAttr;
    char* CmtBuf;
    int CmtBufSize;
    int CmtSize;
    int CmtState;
    char Unused[1024];
  } tHeaderDataExW;

typedef struct {
    char* ArcName;
    int OpenMode;
    int OpenResult;
    char* CmtBuf;
    int CmtBufSize;
    int CmtSize;
    int CmtState;
  } tOpenArchiveData;

typedef struct {
    WCHAR* ArcNameW;
    int OpenMode;
    int OpenResult;
    WCHAR* CmtBufW;
    int CmtBufSize;
    int CmtSize;
    int CmtState;
  } tOpenArchiveDataW;

typedef struct {
	int size;
	DWORD PluginInterfaceVersionLow;
	DWORD PluginInterfaceVersionHi;
	char DefaultIniName[MAX_PATH];
} PackDefaultParamStruct;

/* Definition of callback functions called by the DLL
Ask to swap disk for multi-volume archive */
typedef int (__stdcall *tChangeVolProc)(char *ArcName,int Mode);
typedef int (__stdcall *tChangeVolProcW)(WCHAR *ArcName,int Mode);
/* Notify that data is processed - used for progress dialog */
typedef int (__stdcall *tProcessDataProc)(unsigned char *Addr,int Size);
typedef int (__stdcall *tProcessDataProcW)(WCHAR *Addr,int Size);

typedef struct {
  char ArchiveName[260];
  WCHAR ArchiveNameW[260];
  int ArchiveOpen;
  int ArchivePos;
  DWORD ArchiveSize;
  int ArchiveSizeHigh;
  int ArchiveDate;
  tChangeVolProc ChangeVolProc;
  tChangeVolProcW ChangeVolProcW;
  tProcessDataProc ProcessDataProc;
  tProcessDataProcW ProcessDataProcW;
} ArchiveRec;

int maxarchives=10;

/***************************** Unicode extension **********************************/

#define BZ_HDR_0 0x30   /* '0' */
#define SET_BINARY_MODE(file) setmode(fileno(file),O_BINARY)

BZFILE * BZ2_wbzopen_or_bzdopen
               ( const WCHAR *path,   /* no use when bzdopen */
                 int fd,             /* no use when bzdopen */
                 const char *mode,
                 int open_mode)      /* bzopen: 0, bzdopen:1 */
{
   int    bzerr;
   char   unused[BZ_MAX_UNUSED];
   int    blockSize100k = 9;
   int    writing       = 0;
   WCHAR   mode2[10]     = L"";
   FILE   *fp           = NULL;
   BZFILE *bzfp         = NULL;
   int    verbosity     = 0;
   int    workFactor    = 30;
   int    smallMode     = 0;
   int    nUnused       = 0; 

   if(mode==NULL){return NULL;}
   while(*mode){
      switch(*mode){
      case 'r':
         writing = 0;break;
      case 'w':
         writing = 1;break;
      case 's':
         smallMode = 1;break;
      default:
         if(isdigit(*mode)){
            blockSize100k = *mode-BZ_HDR_0;
         }else{
            /* ignore */
         }
      }
      mode++;
   }
   wcscat(mode2, writing ? L"w" : L"r" );
   wcscat(mode2,L"b");   /* binary mode */

   if(open_mode==0){
      if(path==NULL || path[0]==0){
        fp = (writing ? stdout : stdin);
        SET_BINARY_MODE(fp);
      }else{
        fp = _wfopen(path,mode2);
      }
   }else{
#ifdef BZ_STRICT_ANSI
      fp = NULL;
#else
      fp = _wfdopen(fd,mode2);
#endif
   }
   if(fp==NULL){return NULL;}

   if(writing){
      /* Guard against total chaos and anarchy -- JRS */
      if (blockSize100k < 1) blockSize100k = 1;
      if (blockSize100k > 9) blockSize100k = 9; 
#ifndef disablebzipcompress
      bzfp = BZ2_bzWriteOpen(&bzerr,fp,blockSize100k,verbosity,workFactor);
#endif
   }else{
      bzfp = BZ2_bzReadOpen(&bzerr,fp,verbosity,smallMode,unused,nUnused);
   }
   if(bzfp==NULL){
      if(fp!=stdin && fp!=stdout) fclose(fp);
      return NULL;
   }
   return bzfp;
}

BZFILE * BZ_API(BZ2_wbzopen)
               ( const WCHAR *path,
                 const char *mode )
{
   return BZ2_wbzopen_or_bzdopen(path,-1,mode,/*bzopen*/0);
}

/***************************************************************/

CRITICAL_SECTION slotcriticalsection;
BOOL slotcriticalsectioninitialized=false;

/* Keep a list of currently open archives (up to a maximum of maxarchives) */
static ArchiveRec ArchiveList[11];   /* 0 for packing, 1-10 for unpacking! */
                                      /* Important: array must be initialized to all 0! */

int __stdcall OpenArchive(tOpenArchiveData* ArchiveData)
{
    int ArchiveHandle;
    char sig[8];    /* File signature */
    int n;
    int i;
    int ArcHandle;
    int foundslot;
    unsigned short date, time;
#ifdef _MSC_VER
		struct _stat ftbuf;
		struct tm* fttm;
#endif		
	if (!slotcriticalsectioninitialized) {
		slotcriticalsectioninitialized=true;
		InitializeCriticalSection (&slotcriticalsection);
	}
	EnterCriticalSection (&slotcriticalsection);
    foundslot=0;
    for (i=1;i<=maxarchives;i++)
      if (ArchiveList[i].ArchiveOpen==0) {
        foundslot=1;
        ArcHandle=i;
		ArchiveList[ArcHandle].ArchiveOpen=1;
        break;
      }
    LeaveCriticalSection (&slotcriticalsection);

    if (foundslot==0) {
      ArchiveData->OpenResult=E_NO_MEMORY;
	  ArchiveList[ArcHandle].ArchiveOpen=0;
      return 0;
    }
    if((ArchiveHandle = _open(ArchiveData->ArcName,_O_BINARY | _O_RDONLY))==-1){
      ArchiveData->OpenResult=E_EOPEN;  
      ArchiveList[ArcHandle].ArchiveOpen=0;
      return 0;
    }
    
    n = _read ( ArchiveHandle, sig, 8 );
    if (n != 8) {
      _close(ArchiveHandle);
      ArchiveData->OpenResult=E_EREAD;
	  ArchiveList[ArcHandle].ArchiveOpen=0;
      return 0;
    }
    
    __int64 size64=_lseeki64(ArchiveHandle,0,SEEK_END);
	
	ArchiveList[ArcHandle].ArchiveSize=(DWORD)(size64 & 0xFFFFFFFF);
	ArchiveList[ArcHandle].ArchiveSizeHigh=(DWORD)((size64>>32) & 0xFFFFFFFF);
#ifdef _MSC_VER
		_stat(ArchiveData->ArcName,&ftbuf);
		fttm=localtime(&ftbuf.st_mtime);
		time=fttm->tm_hour<<11 | fttm->tm_min<<5 | fttm->tm_sec;
		date=(fttm->tm_year-80)<<9 | (fttm->tm_mon+1)<<5 | fttm->tm_mday;
#else
    _dos_getftime( ArchiveHandle, &date, &time );
#endif
		ArchiveList[ArcHandle].ArchiveDate=65536*date+time;

    _close(ArchiveHandle);
    if (!((sig[0]=='B') && (sig[1]=='Z') && (sig[2]=='h'))) {
      ArchiveData->OpenResult=E_UNKNOWN_FORMAT;
	  ArchiveList[ArcHandle].ArchiveOpen=0;
      return 0;
    }
    strcpy(ArchiveList[ArcHandle].ArchiveName,ArchiveData->ArcName);
	ArchiveList[ArcHandle].ArchiveNameW[0]=0;
    ArchiveList[ArcHandle].ArchiveOpen=1;
    ArchiveList[ArcHandle].ArchivePos=0;
    return ArcHandle;  /* returns pseudohandle */
}

int __stdcall OpenArchiveW(tOpenArchiveDataW* ArchiveData)
{
    int ArchiveHandle;
    char sig[8];    /* File signature */
    int n;
    int i;
    int ArcHandle;
    int foundslot;
    unsigned short date, time;
#ifdef _MSC_VER
		struct _stat ftbuf;
		struct tm* fttm;
#endif		
	if (!slotcriticalsectioninitialized) {
		slotcriticalsectioninitialized=true;
		InitializeCriticalSection (&slotcriticalsection);
	}
	EnterCriticalSection (&slotcriticalsection);
    foundslot=0;
    for (i=1;i<=maxarchives;i++)
      if (ArchiveList[i].ArchiveOpen==0) {
        foundslot=1;
        ArcHandle=i;
        break;
      }
	LeaveCriticalSection (&slotcriticalsection);

    if (foundslot==0) {
      ArchiveData->OpenResult=E_NO_MEMORY;
      return 0;
    }
    if((ArchiveHandle = _wopen(ArchiveData->ArcNameW,_O_BINARY | _O_RDONLY))==-1){
      ArchiveData->OpenResult=E_EOPEN;  
      return 0;
    }
    
    n = _read ( ArchiveHandle, sig, 8 );
    if (n != 8) {
      _close(ArchiveHandle);
      ArchiveData->OpenResult=E_EREAD;
      return 0;
    }
    
    __int64 size64=_lseeki64(ArchiveHandle,0,SEEK_END);
	
	ArchiveList[ArcHandle].ArchiveSize=(DWORD)(size64 & 0xFFFFFFFF);
	ArchiveList[ArcHandle].ArchiveSizeHigh=(DWORD)((size64>>32) & 0xFFFFFFFF);
#ifdef _MSC_VER
		_wstat(ArchiveData->ArcNameW,&ftbuf);
		fttm=localtime(&ftbuf.st_mtime);
		time=fttm->tm_hour<<11 | fttm->tm_min<<5 | fttm->tm_sec;
		date=(fttm->tm_year-80)<<9 | (fttm->tm_mon+1)<<5 | fttm->tm_mday;
#else
    _dos_getftime( ArchiveHandle, &date, &time );
#endif
		ArchiveList[ArcHandle].ArchiveDate=65536*date+time;

    _close(ArchiveHandle);
    if (!((sig[0]=='B') && (sig[1]=='Z') && (sig[2]=='h'))) {
      ArchiveData->OpenResult=E_UNKNOWN_FORMAT;
      return 0;
    }
	ArchiveList[ArcHandle].ArchiveName[0]=0;
    wcscpy(ArchiveList[ArcHandle].ArchiveNameW,ArchiveData->ArcNameW);
    ArchiveList[ArcHandle].ArchiveOpen=1;
    ArchiveList[ArcHandle].ArchivePos=0;
    return ArcHandle;  /* returns pseudohandle */
}

int __stdcall CloseArchive(int hArcData)
{
    if (hArcData>0 && hArcData<=maxarchives) {
      ArchiveList[hArcData].ArchiveOpen=0;
      return 0;
    } else  
      return E_ECLOSE;
}

int __stdcall ReadHeader(int hArcData,tHeaderData* HeaderData)
{
    if (hArcData>0 && hArcData<=maxarchives && ArchiveList[hArcData].ArchiveOpen) {
        char* p;
        char* p1;                  /* Build file name from archive name without extension */
        if (ArchiveList[hArcData].ArchivePos>0)
          return E_END_ARCHIVE; 
        p=strrchr(ArchiveList[hArcData].ArchiveName,'\\');
        if (p==NULL) p=ArchiveList[hArcData].ArchiveName; else p++;
        p1=strrchr(p,'.');
        if (p1 != NULL) {
          strncpy(HeaderData->FileName,p,p1-p);
        } else {
          strcpy(HeaderData->FileName,p);  
        }
        strcpy(HeaderData->ArcName,ArchiveList[hArcData].ArchiveName);
        HeaderData->PackSize=ArchiveList[hArcData].ArchiveSize;
        HeaderData->UnpSize=ArchiveList[hArcData].ArchiveSize;
        HeaderData->FileTime=ArchiveList[hArcData].ArchiveDate;
        ArchiveList[hArcData].ArchivePos+=1;  /* next file in multifile archive */
        return 0;
    } else {
        return E_EREAD;
    }    
}

int __stdcall ReadHeaderEx(int hArcData,tHeaderDataEx* HeaderDataEx)
{
    if (hArcData>0 && hArcData<=maxarchives && ArchiveList[hArcData].ArchiveOpen) {
        char* p;
        char* p1;                  /* Build file name from archive name without extension */
        if (ArchiveList[hArcData].ArchivePos>0)
          return E_END_ARCHIVE; 
        p=strrchr(ArchiveList[hArcData].ArchiveName,'\\');
        if (p==NULL) p=ArchiveList[hArcData].ArchiveName; else p++;
        p1=strrchr(p,'.');
        if (p1 != NULL) {
          strncpy(HeaderDataEx->FileName,p,p1-p);
        } else {
          strcpy(HeaderDataEx->FileName,p);  
        }
        strcpy(HeaderDataEx->ArcName,ArchiveList[hArcData].ArchiveName);
        HeaderDataEx->PackSize=ArchiveList[hArcData].ArchiveSize;
        HeaderDataEx->PackSizeHigh=ArchiveList[hArcData].ArchiveSizeHigh;
        HeaderDataEx->UnpSize=ArchiveList[hArcData].ArchiveSize;
        HeaderDataEx->UnpSizeHigh=ArchiveList[hArcData].ArchiveSizeHigh;
        HeaderDataEx->FileTime=ArchiveList[hArcData].ArchiveDate;
        ArchiveList[hArcData].ArchivePos+=1;  /* next file in multifile archive */
        return 0;
    } else {
        return E_EREAD;
    }    
}

int __stdcall ReadHeaderExW(int hArcData,tHeaderDataExW* HeaderDataExW)
{
    if (hArcData>0 && hArcData<=maxarchives && ArchiveList[hArcData].ArchiveOpen) {
        WCHAR* p;
        WCHAR* p1;                  /* Build file name from archive name without extension */
        if (ArchiveList[hArcData].ArchivePos>0)
          return E_END_ARCHIVE; 
        p=wcsrchr(ArchiveList[hArcData].ArchiveNameW,'\\');
        if (p==NULL) p=ArchiveList[hArcData].ArchiveNameW; else p++;
        p1=wcsrchr(p,'.');
        if (p1 != NULL) {
          wcsncpy(HeaderDataExW->FileNameW,p,p1-p);
        } else {
          wcscpy(HeaderDataExW->FileNameW,p);  
        }
        wcscpy(HeaderDataExW->ArcNameW,ArchiveList[hArcData].ArchiveNameW);
        HeaderDataExW->PackSize=ArchiveList[hArcData].ArchiveSize;
        HeaderDataExW->PackSizeHigh=ArchiveList[hArcData].ArchiveSizeHigh;
        HeaderDataExW->UnpSize=ArchiveList[hArcData].ArchiveSize;
        HeaderDataExW->UnpSizeHigh=ArchiveList[hArcData].ArchiveSizeHigh;
        HeaderDataExW->FileTime=ArchiveList[hArcData].ArchiveDate;
        ArchiveList[hArcData].ArchivePos+=1;  /* next file in multifile archive */
        return 0;
    } else {
        return E_EREAD;
    }    
}

typedef struct {
     FILE* handle;
} fptr;

typedef fptr* pfptr;

static BOOL myfeof ( FILE* f )
{
   int c = fgetc ( f );
   if (c == EOF) return true;
   ungetc ( c, f );
   return false;
} 

int __stdcall ProcessFile(int hArcData,int Operation,char* DestPath,char* DestName)
{
    if (hArcData<1 || hArcData>maxarchives)
        return E_EREAD;

    if (Operation==PK_SKIP)
        return 0;
        
    if (ArchiveList[hArcData].ArchiveOpen && ArchiveList[hArcData].ArchiveName[0] &&
        ((Operation==PK_EXTRACT && DestName !=NULL) || (Operation==PK_TEST))) {
        int len,i;
        char buff[0x1000];
		unsigned char unused[BZ_MAX_UNUSED]; // data unused in that stream
		void*   unusedTmpV;
		unsigned char*  unusedTmp; 
		int nUnused; 
		BZFILE *BZ2fp_r;
        FILE *fp_w;
		FILE *inStream;
        int aborted;
		int bzerr;
        __int64 oldpos;
        __int64 curpos;
		BOOL smallMode = false;
		int verbosity = 0;
		int streamNo = 0;

		if (ArchiveList[hArcData].ProcessDataProc) {
			if (ArchiveList[hArcData].ProcessDataProc((unsigned char*)(DestName ? DestName:ArchiveList[hArcData].ArchiveName),0)==0)
				return E_EABORTED;
		}
        
		if (Operation==PK_EXTRACT)
                if ((fp_w = fopen(DestName,"wb"))==NULL)
                        return E_ECREATE;

		inStream = fopen ( ArchiveList[hArcData].ArchiveName, "rb" );
		if ( inStream == NULL ) {
			fclose(fp_w);
			return E_EOPEN;
		}

		// new: handle multi-part bzip2 files!
		nUnused=0;
		while (true) { 
			streamNo++;
			BZ2fp_r = BZ2_bzReadOpen ( &bzerr, inStream, verbosity, (int)smallMode, unused, nUnused);
			if (BZ2fp_r == NULL || bzerr != BZ_OK) {
				fclose(fp_w);
				if (streamNo==1)
					return E_EOPEN;
				else
					return 0;  // ignore trailing garbage
			}
			aborted=0;
			oldpos=0;

			bzerr=BZ_OK;
			while (bzerr == BZ_OK && aborted==0) {
				len=BZ2_bzRead(&bzerr,BZ2fp_r,buff,0x1000);
				if ((Operation==PK_EXTRACT) && (bzerr == BZ_OK || bzerr == BZ_STREAM_END) && len > 0) {
					fwrite(buff,1,len,fp_w);
					if (ferror(fp_w)) {
						len=-1;
						break;
					}
				}
				if (Operation==PK_EXTRACT || Operation==PK_TEST) {
					if (ArchiveList[hArcData].ProcessDataProc) {
							curpos=_telli64(((pfptr)BZ2fp_r)->handle->_file);
							if (ArchiveList[hArcData].ProcessDataProc(NULL,(int)(curpos-oldpos))==0)
							aborted=1;
							oldpos=curpos;  
					}
				}
			}
			if (bzerr == BZ_STREAM_END) {
				BZ2_bzReadGetUnused ( &bzerr, BZ2fp_r, &unusedTmpV, &nUnused );
				if (bzerr != BZ_OK)
					len=-1;
				unusedTmp = (unsigned char*)unusedTmpV;
				for (i = 0; i < nUnused; i++) unused[i] = unusedTmp[i]; 
				if (bzerr != BZ_OK)
					len=-1;
				// are we at the end of the archive now?
				if (nUnused == 0 && myfeof(inStream)) {
					len=0;
					break;
				}
			} else {
				BZ2_bzReadClose ( &bzerr, BZ2fp_r ); 
				if (streamNo==1)
					len=-1;
				else
					len=0;
				break;
			}
		}
		fclose (inStream); 

        if (Operation==PK_EXTRACT)
                fclose(fp_w);
        if (aborted)
                return E_EABORTED;
        else if (len<0) {
				if (DestName)
					DeleteFile(DestName);
                return E_BAD_ARCHIVE;
        } else
                return 0;
    } else
        return E_EREAD;
}

int __stdcall ProcessFileW(int hArcData,int Operation,WCHAR* DestPath,WCHAR* DestName)
{
    if (hArcData<1 || hArcData>maxarchives)
        return E_EREAD;

    if (Operation==PK_SKIP)
        return 0;
        
    if (ArchiveList[hArcData].ArchiveOpen && ArchiveList[hArcData].ArchiveNameW[0] &&
        ((Operation==PK_EXTRACT && DestName !=NULL) || (Operation==PK_TEST))) {
        int len,i;
        char buff[0x1000];
		unsigned char unused[BZ_MAX_UNUSED]; // data unused in that stream
		void*   unusedTmpV;
		unsigned char*  unusedTmp; 
		int nUnused; 
		BZFILE *BZ2fp_r;
        FILE *fp_w;
		FILE *inStream;
        int aborted;
		int bzerr;
        __int64 oldpos;
        __int64 curpos;
		BOOL smallMode = false;
		int verbosity = 0;
		int streamNo = 0;

		if (ArchiveList[hArcData].ProcessDataProcW) {
			if (ArchiveList[hArcData].ProcessDataProcW(DestName ? DestName:ArchiveList[hArcData].ArchiveNameW,0)==0)
				return E_EABORTED;
		}
        
		if (Operation==PK_EXTRACT)
                if ((fp_w = _wfopen(DestName,L"wb"))==NULL)
                        return E_ECREATE;

		inStream = _wfopen ( ArchiveList[hArcData].ArchiveNameW, L"rb" );
		if ( inStream == NULL ) {
			fclose(fp_w);
			return E_EOPEN;
		}

		// new: handle multi-part bzip2 files!
		nUnused=0;
		while (true) { 
			streamNo++;
			BZ2fp_r = BZ2_bzReadOpen ( &bzerr, inStream, verbosity, (int)smallMode, unused, nUnused);
			if (BZ2fp_r == NULL || bzerr != BZ_OK) {
				fclose(fp_w);
				if (streamNo==1)
					return E_EOPEN;
				else
					return 0;  // ignore trailing garbage
			}
			aborted=0;
			oldpos=0;

			bzerr=BZ_OK;
			while (bzerr == BZ_OK && aborted==0) {
				len=BZ2_bzRead(&bzerr,BZ2fp_r,buff,0x1000);
				if ((Operation==PK_EXTRACT) && (bzerr == BZ_OK || bzerr == BZ_STREAM_END) && len > 0) {
					fwrite(buff,1,len,fp_w);
					if (ferror(fp_w)) {
						len=-1;
						break;
					}
				}
				if (Operation==PK_EXTRACT || Operation==PK_TEST) {
					if (ArchiveList[hArcData].ProcessDataProc) {
							curpos=_telli64(((pfptr)BZ2fp_r)->handle->_file);
							if (ArchiveList[hArcData].ProcessDataProc(NULL,(int)(curpos-oldpos))==0)
							aborted=1;
							oldpos=curpos;  
					}
				}
			}
			if (bzerr == BZ_STREAM_END) {
				BZ2_bzReadGetUnused ( &bzerr, BZ2fp_r, &unusedTmpV, &nUnused );
				if (bzerr != BZ_OK)
					len=-1;
				unusedTmp = (unsigned char*)unusedTmpV;
				for (i = 0; i < nUnused; i++) unused[i] = unusedTmp[i]; 
				if (bzerr != BZ_OK)
					len=-1;
				// are we at the end of the archive now?
				if (nUnused == 0 && myfeof(inStream)) {
					len=0;
					break;
				}
			} else {
				BZ2_bzReadClose ( &bzerr, BZ2fp_r ); 
				if (streamNo==1)
					len=-1;
				else
					len=0;
				break;
			}
		}
		fclose (inStream); 

        if (Operation==PK_EXTRACT)
                fclose(fp_w);
        if (aborted)
                return E_EABORTED;
        else if (len<0) {
				if (DestName)
					DeleteFileW(DestName);
                return E_BAD_ARCHIVE;
        } else
                return 0;
    } else
        return E_EREAD;
}

void __stdcall SetChangeVolProc(int hArcData,tChangeVolProc pChangeVolProc)
{
    int index=hArcData;
    if (hArcData<1 || hArcData>maxarchives)
      index=0;
    ArchiveList[index].ChangeVolProc=pChangeVolProc;
}

void __stdcall SetProcessDataProc(int hArcData,tProcessDataProc pProcessDataProc)
{
    int index=hArcData;
    if (hArcData<1 || hArcData>maxarchives)
      index=0;
    ArchiveList[index].ProcessDataProc=pProcessDataProc;    
}

void __stdcall SetChangeVolProcW(int hArcData,tChangeVolProcW pChangeVolProcW)
{
    int index=hArcData;
    if (hArcData<1 || hArcData>maxarchives)
      index=0;
    ArchiveList[index].ChangeVolProcW=pChangeVolProcW;
}

void __stdcall SetProcessDataProcW(int hArcData,tProcessDataProcW pProcessDataProcW)
{
    int index=hArcData;
    if (hArcData<1 || hArcData>maxarchives)
      index=0;
    ArchiveList[index].ProcessDataProcW=pProcessDataProcW;
}

int FileExists(char *FileName)
{
    FILE *fp_r;
    int res;
    fp_r = fopen(FileName,"rb");
    res=fp_r!=NULL;
    if (res) fclose(fp_r);
    return res;
}

int FileExistsW(WCHAR *FileName)
{
    FILE *fp_r;
    int res;
    fp_r = _wfopen(FileName,L"rb");
    res=fp_r!=NULL;
    if (res) fclose(fp_r);
    return res;
}

int RenToTemp(char *FileName,char *RenamedName)    /* returns RenamedName */
{
    char *p;
    char ch1;
    char ch2;
    strcpy(RenamedName,FileName);
    p=RenamedName;
    while (p[0]) p++;
    strncat(RenamedName,".#00",259);
    p+=2;
    for (ch1='0';ch1<='9';ch1++) {
      p[0]=ch1;
      for (ch2='0';ch2<='9';ch2++) {
        p[1]=ch2;
        if (!rename(FileName,RenamedName))
          return 1;
      }
    }
    return 0;    /* rename failed! */
}

int RenToTempW(WCHAR *FileName,WCHAR *RenamedName)    /* returns RenamedName */
{
    WCHAR *p;
    WCHAR ch1;
    WCHAR ch2;
    wcscpy(RenamedName,FileName);
    p=RenamedName;
    while (p[0]) p++;
    wcsncat(RenamedName,L".#00",259);
    p+=2;
    for (ch1='0';ch1<='9';ch1++) {
      p[0]=ch1;
      for (ch2='0';ch2<='9';ch2++) {
        p[1]=ch2;
        if (MoveFileW(FileName,RenamedName))
          return 1;
      }
    }
    return 0;    /* rename failed! */
}

/* Functions for packing files */

int __stdcall PackFiles(char *PackedFile,char *SubPath,char *SrcPath,
      char *AddList,int Flags)
{
  char *p;
  int retval=0;

  p=AddList;
  if (!p || !p[0]) return E_NO_FILES;
  while (p[0]) p++;
  p++;
  if (p[0]) return E_TOO_MANY_FILES;
  // Start to pack 
  {
        int len;
        int filethere;
        char buff[0x1000];
        char mode[10];
        char FullName[260];
        char OldName[260];
        BZFILE *BZ2fp_w;
        FILE *fp_r;
        int level = GetPrivateProfileInt("bzip2dll","CompressionRate",6,inifilename);
        if (level<=0) level=1;
        if (level>=9) level=9;

        strcpy(FullName,SrcPath);
        strcat(FullName,AddList);

        filethere=FileExists(PackedFile);
        if (filethere) {
                // Make sure the added filename is the same as the archive name without extension!
                char *p;
                char TestName[260];
                strcpy(TestName,PackedFile);
                p=strrchr(TestName,'.');
                if (p) p[0]='\0';
                p=strrchr(TestName,'\\');
                if (p) p++;
                if (!p || stricmp(p,AddList)!=0)
                        return E_TOO_MANY_FILES;
                filethere=RenToTemp(PackedFile,OldName);
        }        

        if((fp_r = fopen(FullName,"rb"))==NULL) {
                if (filethere)
                        rename(OldName,PackedFile);
                return E_EOPEN;
        }        

        mode[0]='w';
        mode[1] = '0'+level;
        mode[2] = '\0';

        if ((BZ2fp_w = BZ2_bzopen(PackedFile,mode))==NULL) {
                fclose(fp_r);
                if (filethere)
                        rename(OldName,PackedFile);
                return E_ECREATE;
        }
        while((len=fread(buff,1,0x1000,fp_r))>0 && retval==0) {
                if (ArchiveList[0].ProcessDataProc)
                        if (ArchiveList[0].ProcessDataProc((unsigned char*)AddList,len)==0) {
                                retval=E_EABORTED;
                                break;
                        }        
                if (BZ2_bzwrite(BZ2fp_w,buff,len)==-1) {
                        retval=E_EWRITE;
                        break;
                }
        }
        BZ2_bzclose(BZ2fp_w);
        fclose(fp_r);
        // Error occurred -> delete new archive and restore old (if exists)
        if (retval) {
                remove(PackedFile);
                if (filethere)
                        rename(OldName,PackedFile);
        } else {
                if (filethere)
                        unlink(OldName);
                // Move to archive - only if packing was successful!
                if (Flags & PK_PACK_MOVE_FILES)
                        unlink(FullName);
        }
                
  }
  return retval; 
}

int __stdcall PackFilesW(WCHAR *PackedFile,WCHAR *SubPath,WCHAR *SrcPath,
      WCHAR *AddList,int Flags)
{
  WCHAR *p;
  int retval=0;

  p=AddList;
  if (!p || !p[0]) return E_NO_FILES;
  while (p[0]) p++;
  p++;
  if (p[0]) return E_TOO_MANY_FILES;
  // Start to pack 
  {
        int len;
        int filethere;
        char buff[0x1000];
        char mode[10];
        WCHAR FullName[MAX_PATH];
        WCHAR OldName[MAX_PATH];
        BZFILE *BZ2fp_w;
        FILE *fp_r;
        int level = GetPrivateProfileInt("bzip2dll","CompressionRate",6,inifilename);
        if (level<=0) level=1;
        if (level>=9) level=9;

        wcscpy(FullName,SrcPath);
        wcsncat(FullName,AddList,MAX_PATH);

        filethere=FileExistsW(PackedFile);
        if (filethere) {
                // Make sure the added filename is the same as the archive name without extension!
                WCHAR *p;
                WCHAR TestName[260];
                wcscpy(TestName,PackedFile);
                p=wcsrchr(TestName,'.');
                if (p) p[0]='\0';
                p=wcsrchr(TestName,'\\');
                if (p) p++;
                if (!p || wcsicmp(p,AddList)!=0)
                        return E_TOO_MANY_FILES;
                filethere=RenToTempW(PackedFile,OldName);
        }        
 
        if((fp_r = _wfopen(FullName,L"rb"))==NULL) {
                if (filethere)
                        MoveFileW(OldName,PackedFile);
                return E_EOPEN;
        }        

        mode[0]='w';
        mode[1] = '0'+level;
        mode[2] = '\0';

        if ((BZ2fp_w = BZ2_wbzopen(PackedFile,mode))==NULL) {
                fclose(fp_r);
                if (filethere)
                        MoveFileW(OldName,PackedFile);
                return E_ECREATE;
        }
        while((len=fread(buff,1,0x1000,fp_r))>0 && retval==0) {
                if (ArchiveList[0].ProcessDataProcW)
                        if (ArchiveList[0].ProcessDataProcW((WCHAR*)AddList,len)==0) {
                                retval=E_EABORTED;
                                break;
                        }        
                if (BZ2_bzwrite(BZ2fp_w,buff,len)==-1) {
                        retval=E_EWRITE;
                        break;
                }
        }
        BZ2_bzclose(BZ2fp_w);
        fclose(fp_r);
        // Error occurred -> delete new archive and restore old (if exists)
        if (retval) {
                DeleteFileW(PackedFile);
                if (filethere)
                        MoveFileW(OldName,PackedFile);
        } else {
                if (filethere)
                        DeleteFileW(OldName);
                // Move to archive - only if packing was successful!
                if (Flags & PK_PACK_MOVE_FILES)
                        DeleteFileW(FullName);
        }
                
  }
  return retval; 
}

int __stdcall DeleteFiles(char *PackedFile,char *DeleteList)
{
  // Since this archive contains only one file anyway, just delete the whole archive!
  if (DeleteList && DeleteList[0])
    if (unlink(PackedFile))
      return E_EWRITE;
    else
      return 0;
  return E_BAD_DATA;
}

int __stdcall DeleteFilesW(WCHAR *PackedFile,WCHAR *DeleteList)
{
  // Since this archive contains only one file anyway, just delete the whole archive!
  if (DeleteList && DeleteList[0])
    if (!DeleteFileW(PackedFile))
      return E_EWRITE;
    else
      return 0;
  return E_BAD_DATA;
}

int __stdcall GetPackerCaps()
{
  return PK_CAPS_NEW | PK_CAPS_MODIFY | PK_CAPS_DELETE | PK_CAPS_OPTIONS | PK_CAPS_MEMPACK | PK_CAPS_BY_CONTENT | PK_CAPS_SEARCHTEXT;
}

HANDLE hBrush;

BOOL CALLBACK DialogProc(HWND hwndDlg,UINT  uMsg,WPARAM wParam,LPARAM  lParam)
{
   int i;
   char num[4];
   char ch;
   num[1]='\0';
   switch (uMsg) {
     case WM_INITDIALOG:
       for (ch='1';ch<='9';ch++) {
           num[0]=ch;
           SendDlgItemMessage(hwndDlg,103,CB_ADDSTRING,0,(LPARAM)&num);
       }
       i=GetPrivateProfileInt("bzip2dll","CompressionRate",6,inifilename);
       SendDlgItemMessage(hwndDlg,103,CB_SETCURSEL,i-1,0);
       hBrush=CreateSolidBrush(GetSysColor(COLOR_BTNFACE));
       return TRUE;
       break;
     case WM_CTLCOLORDLG:
     case WM_CTLCOLORSTATIC:
     case WM_CTLCOLORBTN:
       SetBkColor((HDC)wParam,GetSysColor(COLOR_BTNFACE));
       SetTextColor((HDC)wParam,GetSysColor(COLOR_BTNTEXT));
       return (BOOL)hBrush;
       break;
     case WM_COMMAND:
       switch (wParam) {
         case 104:    /* OK  */
           num[0]=(char)('1'+SendDlgItemMessage(hwndDlg,103,CB_GETCURSEL,0,0));
           WritePrivateProfileString("bzip2dll","CompressionRate",num,inifilename);
           DeleteObject(hBrush);
           EndDialog(hwndDlg,1);
         case 2:    /* Cancel */
           DeleteObject(hBrush);
           EndDialog(hwndDlg,0);
       }
     break;
   }
   return FALSE;
}

void __stdcall ConfigurePacker(HWND ParentHandle, HINSTANCE hinstance)
{	
   /* You may ask for packing (NOT unpacking!) parameters here */
   DialogBox(hinstance,"CONFDLG",ParentHandle,(DLGPROC)&DialogProc);
}

typedef struct {
	bz_stream strm;
	char filename[MAX_PATH];
	bool needsendheader;
} tmempackstruct,*pmempackstruct;

char* strlcpy(char* p,char*p2,int maxlen)
{
	if ((int)strlen(p2)>=maxlen) {
		strncpy(p,p2,maxlen);
		p[maxlen]=0;
	} else
		strcpy(p,p2);
	return p;
}

HANDLE __stdcall StartMemPack(int options,char* filename)
{
	int ret;
	pmempackstruct pmps;
	int blockSize100k;
	int verbosity     = 0;
	int workFactor    = 30;
	
	pmps=(pmempackstruct)malloc(sizeof(tmempackstruct));
	if (!pmps) return 0;
	memset(pmps,0,sizeof(tmempackstruct));
	strlcpy(pmps->filename,filename,MAX_PATH-1);
	pmps->needsendheader=(options & MEM_OPTIONS_WANTHEADERS)!=0;
	if (!pmps->needsendheader) {    // can't create data without header!
		free(pmps);
		return 0;
	}

	blockSize100k=GetPrivateProfileInt("bzip2dll","CompressionRate",6,inifilename);
	ret = BZ2_bzCompressInit ( &pmps->strm, blockSize100k, 
                          verbosity, workFactor );
	if (ret != BZ_OK) {
		free(pmps);
		return 0;
	}
 
	return (HANDLE)pmps;
}

int __stdcall PackToMem(HANDLE hMemPack,char* BufIn,int InLen,int* Taken,
      char* BufOut,int OutLen,int* Written,int* SeekBy)
{
	pmempackstruct pmps;
	int ret;
	__int64 last_totalin,last_totalout,new_totalin,new_totalout;

	pmps=(pmempackstruct)hMemPack;

	last_totalin=pmps->strm.total_in_hi32;
	last_totalin=(last_totalin<<32) | pmps->strm.total_in_lo32;
	last_totalout=pmps->strm.total_out_hi32;
	last_totalout=(last_totalout<<32) | pmps->strm.total_out_lo32;
	pmps->strm.next_in = BufIn;
	pmps->strm.next_out = BufOut;
	pmps->strm.avail_in = InLen;
	pmps->strm.avail_out = OutLen;

	if (SeekBy)
		*SeekBy=0;    // no need to seek to any different position in the file

	if (InLen)
		ret = BZ2_bzCompress ( &pmps->strm, BZ_RUN );
	else
		ret = BZ2_bzCompress ( &pmps->strm, BZ_FINISH );

	if (Taken) {
		new_totalin=pmps->strm.total_in_hi32;
		new_totalin=(new_totalin<<32) | pmps->strm.total_in_lo32;
		*Taken=(int)(new_totalin-last_totalin);
	}
	if (Written) {
		new_totalout=pmps->strm.total_out_hi32;
		new_totalout=(new_totalout<<32) | pmps->strm.total_out_lo32;
		*Written=(int)(new_totalout-last_totalout);
	}
	
	if (ret==BZ_STREAM_END) {
//		*SeekBy=-last_totalout; could be used to change the header, e.g. to
//		insert the total size!
		return MEMPACK_DONE;
	}
	else if (ret==BZ_OK || ret==BZ_RUN_OK || ret==BZ_FINISH_OK)
		return MEMPACK_OK;
	else
		return MEMPACK_ERR;
}

int __stdcall DoneMemPack(HANDLE hMemPack)
{
	pmempackstruct pmps;
	if (hMemPack) {
		pmps=(pmempackstruct)hMemPack;
		BZ2_bzCompressEnd(&pmps->strm);
		free((pmempackstruct)hMemPack);
	}
	return 0;
}

int __stdcall GetBackgroundFlags(void)
{
	return BACKGROUND_UNPACK | BACKGROUND_PACK | BACKGROUND_MEMPACK;   // Support both pack+unpack+pack in memory (for tar.bz2)
}

BOOL __stdcall CanYouHandleThisFile(char* filename)
{
    FILE* ArchiveHandle;
    char sig[8];    // File signature
	int n;
	if (filename==NULL)
		return false;

	if (filename[0]==0)
		return false;

    if((ArchiveHandle = fopen(filename,"rb"))==NULL)
      return false;
    
    n = fread ( sig, 1, 8, ArchiveHandle );
    if (n != 8) {
      fclose(ArchiveHandle);
      return false;
    }
    fclose(ArchiveHandle);
    if (!((sig[0]=='B') && (sig[1]=='Z') && (sig[2]=='h'))) {
      return false;
    }
    return true;
}

BOOL __stdcall CanYouHandleThisFileW(WCHAR* filename)
{
    FILE* ArchiveHandle;
    char sig[8];    // File signature
	int n;
	if (filename==NULL)
		return false;

	if (filename[0]==0)
		return false;

    if((ArchiveHandle = _wfopen(filename,L"rb"))==NULL)
      return false;
    
    n = fread ( sig, 1, 8, ArchiveHandle );
    if (n != 8) {
      fclose(ArchiveHandle);
      return false;
    }
    fclose(ArchiveHandle);
    if (!((sig[0]=='B') && (sig[1]=='Z') && (sig[2]=='h'))) {
      return false;
    }
    return true;
}

void __stdcall PackSetDefaultParams(PackDefaultParamStruct* dps)
{
	strlcpy(inifilename,dps->DefaultIniName,MAX_PATH-1);
}
